/*
 *  Copyright (C) 2013-2014 Katharina Breininger
 *
 *   This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.WindowManager;
import ij.gui.GenericDialog;
import ij.gui.Line;
import ij.gui.PolygonRoi;
import ij.gui.Roi;
import ij.measure.Calibration;
import ij.plugin.filter.PlugInFilter;
import ij.plugin.frame.RoiManager;
import ij.process.ImageProcessor;

import java.awt.Polygon;
import java.util.LinkedList;


public class Scaler_ implements PlugInFilter {
	boolean debug = false;
	ImagePlus imp;
	String title;
	int interpolationMethod = ImageProcessor.BILINEAR;
	double scaleFactor;
	Integer[] indices;
	boolean scaleImage;
	boolean scaleX;
	boolean scaleY;
	@Override
	public int setup(String arg, ImagePlus imp) {
		this.imp = imp;
		return DOES_8G + DOES_16 + DOES_32 + DOES_RGB + NO_CHANGES;
	}

	@Override
	public void run(ImageProcessor ip) {
		if(showDialog()) {
			if(indices != null)
				scaleChosenRois();
			if(scaleImage) {
				ImagePlus scaledImg = Scaler_.createRescaledStack(imp, scaleFactor, interpolationMethod, title, true, scaleX, scaleY);
				scaledImg.show();
			}
		}		
	}

	private boolean showDialog() {
		scaleFactor = 10.0;
		String[] methods = ImageProcessor.getInterpolationMethods();
		title = WindowManager.getUniqueName(imp.getTitle());

		GenericDialog gd = new GenericDialog("Scaler plugin options");
		gd.addStringField("Scale factor", "" + scaleFactor);
		gd.addStringField("Title", title, 12);
		gd.addChoice("Interpolation method", methods, methods[interpolationMethod]);
		gd.addStringField("ROIs (RoiManager-indices) you want to scale (start with 0, separate by ,)", "", 24);
		gd.addCheckbox("Scale in x-direction", true);
		gd.addCheckbox("Scale in y-direction", true);
		gd.addCheckbox("Scale image as well", true);
		gd.showDialog();
		if (gd.wasCanceled())
			return false;

		String scaleStr = gd.getNextString(); // out here to not introduce funny behavior if further textfields are added
		title = gd.getNextString();
		interpolationMethod = gd.getNextChoiceIndex();
		String roiString = gd.getNextString();
		scaleX = gd.getNextBoolean();
		scaleY = gd.getNextBoolean();
		scaleImage = gd.getNextBoolean();
		try {
			scaleFactor = Double.parseDouble(scaleStr);
		} catch(NumberFormatException e) {
			IJ.error("Invalid input for scale factor (no number) - plugin aborted.");
			return false;
		}
		if(scaleFactor <= 0) {
			IJ.error("Invalid input for scale factor (negative number) - plugin aborted.");
			return false;
		}
		if(!roiString.equals("")) {
			indices = parseRoiIDs(roiString);
		}
		else indices = null;
		return true;
	}

	public Integer[] parseRoiIDs(String roiString) {
		LinkedList<Integer> roiIndices = new LinkedList<Integer>();
		roiString = roiString.replaceAll("\\s", "");
		String[] numbers = roiString.split(",");
		
		for(int i = 0; i < numbers.length; i++) {
			Integer curInt = null;
			try {
				curInt = Integer.parseInt(numbers[i]);
			}
			catch(NumberFormatException e) {
				IJ.log("Not able to parse "+ numbers[i]);
				continue;
			}
			roiIndices.add(curInt);
			
		}
		Integer[] indices = new Integer[roiIndices.size()];
		roiIndices.toArray(indices);
		if(debug) {
			for(int i = 0; i < indices.length; i++)
				System.out.println(indices[i]);
		}	
		return indices;
	}
	
	private void scaleChosenRois() {
		RoiManager roiManager = null;
		if(RoiManager.getInstance() == null)
			roiManager = new RoiManager();
		else 
			roiManager = RoiManager.getInstance();
		Roi[] currentRois = roiManager.getRoisAsArray();
		Roi[] toscale = new Roi[indices.length];
		for(int i = 0; i < indices.length; i++) {
			toscale[i] = currentRois[indices[i]];
			Roi old = currentRois[indices[i]];
			Roi neu = scaleRoi(old, scaleFactor, scaleX, scaleY);
			neu.setName(old.getName()+"-scaled");
			roiManager.addRoi(neu);
		}

	}
	
	/**
	 * This method creates a new, rescaled version of a given image.
	 * @param imp The basis image.
	 * @param scaleFactor The factor by which the image shall be scaled
	 * @param title The title of the new image.
	 * @param statusMessage Flag if a status message is supposed to be shown in the ImageJ status bar.
	 * @return The rescaled image.
	 */
	static ImagePlus createRescaledStack(ImagePlus imp, double scaleFactor, int interpolationMethod, String title, boolean statusMessage) {
		return createRescaledStack(imp, scaleFactor, interpolationMethod, title, statusMessage, true, true);
	}

	/**
	 * This method creates a new, rescaled version of a given image.
	 * @param imp The basis image.
	 * @param scaleFactor The factor by which the image shall be scaled
	 * @param title The title of the new image.
	 * @param statusMessage Flag if a status message is supposed to be shown in the ImageJ status bar.
	 * @return The rescaled image.
	 */
	static ImagePlus createRescaledStack(ImagePlus imp, double scaleFactor, int interpolationMethod, String title, boolean statusMessage, boolean scaleX, boolean scaleY) {
		int nSlices = imp.getStackSize();
		int width=imp.getWidth(), height=imp.getHeight();
		int newWidth = scaleX ? scale(width, scaleFactor) : width; // not sure if rounding is the right way to get new image size
		int newHeight = scaleY ? scale(height, scaleFactor) : height;
		ImagePlus imp2 = imp.createImagePlus();
		ImageStack stack1 = imp.getStack();
		ImageStack stack2 = new ImageStack(newWidth, newHeight);
		ImageProcessor ip1, ip2;
		int method = interpolationMethod;
		if (width==1 || height==1)
			method = ImageProcessor.NONE;
	
		for (int i=1; i<=nSlices; i++) {
			if(statusMessage)
				IJ.showStatus("Scale: " + i + "/" + nSlices);
			ip1 = stack1.getProcessor(i);
			String label = stack1.getSliceLabel(i);
			int interpolSave = ip1.getInterpolationMethod();
			ip1.setInterpolationMethod(method);
			ip2 = ip1.resize(newWidth, newHeight, true);
			ip1.setInterpolationMethod(interpolSave);
			if (ip2!=null)
				stack2.addSlice(label, ip2);
			if(statusMessage)
				IJ.showProgress(i, nSlices);
		}
	
		imp2.setStack(title, stack2);
		Calibration cal = imp2.getCalibration();
		if (cal.scaled()) {
			cal.pixelWidth *= 1.0/scaleFactor;
			cal.pixelHeight *= 1.0/scaleFactor;
		}
		if(statusMessage)
			IJ.showProgress(1.0);
		int[] dim = imp.getDimensions();
		imp2.setDimensions(dim[2], dim[3], dim[4]);
		if (imp.isHyperStack())
			imp2.setOpenAsHyperStack(true);
		return imp2;
	}
	
	/**
	 * This method scales an integer value x by a double factor and rounds it to the closest integer value.
	 */
	private static int scale(int x, double factor) {
		return (int)Math.round(x*factor);
	}

	/**
	 * This method scales a given region of interest (roi) by a double factor.
	 * @param roi The roi that is scaled.
	 * @param factor The roi is scaled by this factor.
	 * @return The scaled region of interest.
	 */
	
	public static Roi scaleRoi(Roi roi, double factor, boolean scaleX, boolean scaleY) {
		switch(roi.getType()) {
		case Roi.LINE:
			return scaleRoi((Line)roi, factor, scaleX, scaleY);
		case Roi.POLYLINE:
			return scaleRoi((PolygonRoi)roi, factor,scaleX, scaleY);
		default:
			throw new IllegalArgumentException("Roitype is not supported");
		}
	}
	
	
	public static Line scaleRoi(Line roi, double factor, boolean scaleX, boolean scaleY) {
		int x1 = roi.x1;
		int x2 = roi.x2;
		int y1 = roi.y1;
		int y2 = roi.y2;
		if(scaleX) {
			x1 = scale(x1, factor);
			x2 = scale(x2, factor);
		}
		if(scaleY) {
			y1 = scale(y1, factor);
			y2 = scale(y2, factor);
		}
		return new Line(x1, y1, x2, y2);
	}

	public static PolygonRoi scaleRoi(PolygonRoi roi, double factor, boolean scaleX, boolean scaleY) {
		if(roi.getType() != Roi.POLYLINE)
			throw new IllegalArgumentException("Only Polylines can be scaled right now!");
		Polygon poly = roi.getPolygon();
		int[] xvals = poly.xpoints;
		int[] yvals = poly.ypoints;
		int[] xvalsnew = new int[poly.npoints];
		int[] yvalsnew = new int[poly.npoints];
		for(int i = 0; i < poly.npoints; i++) {
			xvalsnew[i] = scaleX ? scale(xvals[i], factor): xvals[i];
			yvalsnew[i] = scaleY ? scale(yvals[i], factor): yvals[i];
		}
		return new PolygonRoi(xvalsnew, yvalsnew, yvalsnew.length, Roi.POLYLINE);
	}
}
