/*
 *  Copyright (C) 2013-2014 Katharina Breininger
 *
 *   This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * This class provides an optimization function for the JPOP framework ({@link http://www5.cs.fau.de/research/software/java-parallel-optimization-package/} or 
 * {@link http://peaks.informatik.uni-erlangen.de/jpop-api/overview-summary.html}.
 * <p>
 * The goal is to model a given set of values with a sum of three Gaussian functions and optimize the parameters of these Gaussian functions via
 * the sum of squared differences.
 * The following parameters are optimized: Mean (\mu_i), standard deviation (\sigma_i).
 * <p>
 * <b>Formula: Sum of three Gaussians (in latex):</b><br>
 * g(x) = \sum_{k=1}^{3} \mathcal{N}(x; \mu_k, \sigma_k)  <br>
 * <b>Formula: Optimization function: </b> <br>
 * h(\theta) = \sum_{i = 1}^{N} (f(x_i)-g_s(x_i))^2 --> minimum <br>
 * with \theta = (\mu_1, \sigma_1, \mu_2, \sigma_2, \mu_3, \sigma_3)
 * <p>
 * Except not noted otherwise in the method description, these parameters must have the same order as shown in the line above and will have the same order in the array that contains the final result of the optimization process.
 * <p>
 * Note:
 * For further information please refer to the superclass ScalableGaussianOptimizationFunction. 
 * This implementation is strongly based on the superclass - for the variable scaling factors a constant value of 1 is passed.
 */
public class SimpleGaussianOptimizationFunction extends ScalableGaussianOptimizationFunction {
	double[] yvalues;
	final static double[] scalefactors = {1, 1, 1};
	
	/**
	 * @see ScalableGaussianOptimizationFunction
	 */
	public SimpleGaussianOptimizationFunction(double[] xvalues, double[] yvalues) {
		super(xvalues, yvalues);
	}
		

	/**
	 * Unscaled version.
	 * @param params = {\mu_1, \sigma_1, \mu_2, \sigma_2, \mu_3, \sigma_3}
	 * @param block 
	 */
	public double evaluate(double[] params, int block) {
		if(params.length != 6) {
			throw new IllegalArgumentException("Wrong number of parameters for evaluation!");
		}
		return evaluate(params, scalefactors, block);
	}
	
	/**
	 * Unscaled version.
	 * @param params = {\mu_1, \sigma_1, \mu_2, \sigma_2, \mu_3, \sigma_3}
	 * @param block 
	 */
	static double evaluateTripleGaussian(double x, double[] params) {
		if(params.length != 6) {
			throw new IllegalArgumentException("Wrong number of paramters for gradient calculation!");
		}
		return evaluateTripleGaussian(x, params, scalefactors);
	}

	@Override
	public double[] gradient(double[] params, int block) {
		double[] result = new double[6];
		for(int j = 0; j < 3; ++j) {
			result[j*2] = calculateGradientRespToMean(params, scalefactors, j);
			result[j*2+1] = calculateGradientRespToStandardDev(params, scalefactors, j);
		}
		return result;
	}

	@Override
	public double evaluateBasisFunction(double x, double[] params) {
		if(params.length != 6) {
			throw new IllegalArgumentException("Wrong number of paramters for gradient calculation!");
		}
		return evaluateTripleGaussian(x, params);
	}
}
